#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : DNN
# @Time         : 2020-03-13 13:42
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf


class DNN(tf.keras.layers.Layer):

    def __init__(self,
                 hidden_units_list: List[int]= (64, 32, 16, 4),
                 activation='relu',
                 kernel_regularizer=tf.keras.regularizers.l2(),
                 use_bn=False,
                 dropout_rate=0,
                 seed=666,
                 name='DNN',
                 **kwargs):
        super().__init__(name=name, **kwargs)

        self.hidden_units_list = hidden_units_list
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.kernel_regularizer = kernel_regularizer
        self.use_bn = use_bn
        self.seed = seed
        self.num_layer = len(self.hidden_units_list)

    def build(self, input_shape):
        super().build(input_shape)  # self.built = True

        self.dense_layers = []
        for index, units in enumerate(self.hidden_units_list):
            _ = tf.keras.layers.Dense(
                units,
                activation=self.activation,
                use_bias=True,
                kernel_initializer='glorot_uniform',
                bias_initializer='zeros',
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=None,
                name=f"dense{index}"
            )
            self.dense_layers.append(_)

        # BN
        if self.use_bn:
            self.bn_layers = self.num_layer * [tf.keras.layers.BatchNormalization()]

        self.dropout_layers = self.num_layer * [
            tf.keras.layers.Dropout(self.dropout_rate, seed=self.seed + 666)]  # seed + i

    def call(self, inputs, **kwargs):
        deep_input = inputs
        for i in range(self.num_layer):
            fc = self.dense_layers[i](deep_input)  # 注意下次循环的输入
            fc = self.bn_layers[i](fc) if self.use_bn else fc
            fc = self.dropout_layers[i](fc)
            deep_input = fc

        return fc

    def compute_output_shape(self, input_shape):
        return input_shape[:-1] + (self.hidden_units_list[-1],)

    def get_config(self):
        base_config = super().get_config()
        config = {
            'hidden_units': self.hidden_units_list,
            'activation': self.activation,
            'kernel_regularizer': self.kernel_regularizer,
            'use_bn': self.use_bn,
            'dropout_rate': self.dropout_rate,
            'seed': self.seed
        }
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : PredictionLayer
# @Time         : 2020/4/29 6:07 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf
from tensorflow.keras.initializers import Zeros


class Prediction(tf.keras.layers.Layer):

    def __init__(self, task='binary', use_bias=True, num_class=0, name='Prediction', **kwargs):
        """https://github.com/shenweichen/DeepCTR/issues/211

        :param task: ["binary", "multiclass", "regression"]
        :param use_bias:
        :param kwargs:
        """
        super().__init__(name=name, **kwargs)

        if task not in ["binary", "multiclass", "regression"]:
            raise ValueError("task must be binary, multiclass or regression")

        self.task = task
        self.use_bias = use_bias
        self.num_class = num_class

    def build(self, input_shape):
        super().build(input_shape)

        if self.use_bias:
            self.global_bias = self.add_weight(
                shape=(self.num_class,) if self.task == "multiclass" else (1,),
                initializer=Zeros(), name="global_bias")

    def call(self, inputs, **kwargs):
        x = inputs

        if self.use_bias:
            x = tf.nn.bias_add(x, self.global_bias, data_format='NHWC')

        if self.task == "binary":
            x = tf.sigmoid(x)  # [[pred1], [pred2]]
            output = tf.reshape(x, (-1, 1))
        elif self.task == "multiclass":
            x = tf.nn.softmax(x)
            output = tf.reshape(x, (-1, self.num_class))
        else:
            # regression
            output = x

        return output

    def compute_output_shape(self, input_shape):
        if self.task == "multiclass":
            return (None, self.num_class)
        else:
            return (None, 1)

    def get_config(self, ):
        base_config = super().get_config()
        config = {'task': self.task, 'use_bias': self.use_bias, 'num_class': self.num_class}
        return {**base_config, **config}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : inn.
# @File         : BaseModel
# @Time         : 2020/5/19 2:24 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union

import tensorflow as tf
from tensorflow.python.feature_column.feature_column_v2 import \
    FeatureColumn, NumericColumn, CategoricalColumn, SequenceCategoricalColumn, EmbeddingColumn, \
    BucketizedColumn


class BaseModel(object):

    def __init__(self,  # todo: 初始化基础属性
                 feature_columns: List[FeatureColumn] = None,
                 task='binary',
                 num_class=2,
                 early_stopping_epochs=3):

        self.feature_columns = feature_columns
        self.task = task
        self.num_class = num_class
        self.early_stopping_epochs = early_stopping_epochs

    def plot(self, to_file='model.png',
             show_shapes=False,
             show_layer_names=True,
             rankdir='TB',
             expand_nested=False,
             dpi=96):
        tf.keras.utils.plot_model(self.model, to_file=to_file,
                                  show_shapes=show_shapes,
                                  show_layer_names=show_layer_names,
                                  rankdir=rankdir,
                                  expand_nested=expand_nested,
                                  dpi=dpi)

    def callbacks(self):
        callbacks_list = [
            tf.keras.callbacks.ReduceLROnPlateau(factor=0.9, patience=2, verbose=1, min_lr=0.0001),
            # annealer = LearningRateScheduler(lambda x: min(0.01 * 0.9 ** x, 0.001), verbose=1)

            tf.keras.callbacks.ModelCheckpoint("filepath",
                                               monitor='val_loss',
                                               verbose=0,
                                               save_best_only=True,
                                               save_weights_only=False,
                                               mode='auto',
                                               save_freq='epoch'),

            tf.keras.callbacks.EarlyStopping(monitor='val_loss',
                                             min_delta=0,
                                             patience=self.early_stopping_epochs,
                                             verbose=0,
                                             mode='auto',
                                             baseline=None,
                                             restore_best_weights=False)
        ]
        return callbacks_list

    # todo: 学习率clr_callback, WandbCallback



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepNN.
# @File         : numeric_column
# @Time         : 2020/4/13 4:52 下午
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

import tensorflow as tf


def get_feature_columns(numeric_cols: list, shape=(1,), normalizer_fn=None):
    feature_columns = []
    feature_layer_inputs = {}

    for feat in numeric_cols:
        feature_column = tf.feature_column.numeric_column(feat, shape, normalizer_fn=normalizer_fn)
        feature_columns.append(feature_column)
        feature_layer_inputs[feat] = tf.keras.Input(feature_column.shape, name=feat)
    return feature_columns, feature_layer_inputs
